/*
 * XMOSlisp is a tiny lisp interpreter for XMOS devices.
 * Copyright (C) 2005-2010  David Boozer
 * Copyright (C) 2010       Ruben Vandamme
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "garbage.h"
#include "xmos.h"
#include "xs1.h"

typedef struct obj {
  enum type_code type;
} obj;

typedef obj *(cfunc_type)(obj *);

typedef struct {
  enum type_code type;
  obj *car;
  obj *cdr;
} obj_cons;

typedef struct {
  enum type_code type;
} obj_timer;

typedef struct {
  enum type_code type;
  port name;
} obj_port;

typedef struct {
  enum type_code type;
  int value;
} obj_int;

typedef struct {
  enum type_code type;
  char *name;
} obj_sym;

typedef struct {
  enum type_code type;
  cfunc_type *cfunc;
} obj_prim;

typedef struct {
  enum type_code type;
  obj *code;
  obj *syms;
  obj *env;
} obj_comp;

typedef struct {
  enum type_code type;
  obj *func;
} obj_spec;

/* lists */
obj *names_head, *top_env;

/* quote */
obj *s_quote;

/* () and #t */
obj *nil, *tee;

int num_nodes = 0;
int flag_trace = 0;

#define car(X)                (((obj_cons *) (X))->car)
#define cdr(X)                (((obj_cons *) (X))->cdr)

#define second(X)             (car(cdr(X)))
#define third(X)              (car(cdr(cdr(X))))
#define fourth(X)             (car(cdr(cdr(cdr(X)))))

#define int_value(X)          (((obj_int *) (X))->value)
#define sym_name(X)           (((obj_sym *) (X))->name)
#define prim_cfunc(X)         (((obj_prim *) (X))->cfunc)
#define comp_code(X)          (((obj_comp *) (X))->code)
#define comp_syms(X)          (((obj_comp *) (X))->syms)
#define comp_env(X)           (((obj_comp *) (X))->env)
#define spec_func(X)          (((obj_spec *) (X))->func)
#define port_name(X)          (((obj_port *) (X))->name)

/* function prototypes */

void write (FILE *output, obj *sexp);

obj *read_sexp ();
obj *read_list ();

obj *eval_list (obj *exps, obj *env);
obj *apply (obj *func, obj *vals);

/* construct lisp objects */

obj *cons (obj *first, obj *second) {
  obj_cons *ptr;

  num_nodes++;
  ptr = (obj_cons *) CollectMalloc (sizeof (obj_cons), 2);
  ptr->type = CONS;
  ptr->car = first;
  ptr->cdr = second;
  return (obj *) ptr;
}

obj *make_port (unsigned name) {
  obj_port *ptr;

  num_nodes++;
  ptr = (obj_port *) CollectMalloc(sizeof (obj_port), 0);
  ptr->type = PORT;
  ptr->name = name;
  return (obj *) ptr;
}

obj *make_timer () {
  obj_timer *ptr;

  num_nodes++;
  ptr = (obj_timer *) CollectMalloc (sizeof (obj_timer), 0);
  ptr->type = TIMER;
  //ptr->name = name; // Figure out how to return a timer
  return (obj *) ptr;
}

obj *make_int (int value) {
  obj_int *ptr;

  num_nodes++;
  ptr = (obj_int *) CollectMalloc (sizeof (obj_int), 0);
  ptr->type = INT;
  ptr->value = value;
  return (obj *) ptr;
}

obj *make_sym (char *name) {
  obj_sym *ptr;

  num_nodes++;
  ptr = (obj_sym *) CollectMalloc (sizeof (obj_sym), 0);
  ptr->type = SYM;
  ptr->name = name;
  return (obj *) ptr;
}

obj *make_prim (cfunc_type *cfunc) {
  obj_prim *ptr;

  num_nodes++;
  ptr = (obj_prim *) CollectMalloc (sizeof (obj_prim), 0);
  ptr->type = PRIM;
  ptr->cfunc = cfunc;
  return (obj *) ptr;
}

obj *make_comp (obj *code, obj *syms, obj *env) {
  obj_comp *ptr;

  num_nodes++;
  ptr = (obj_comp *) CollectMalloc (sizeof (obj_comp), 3);
  ptr->type = COMP;
  ptr->code = code;
  ptr->syms = syms;
  ptr->env = env;
  return (obj *) ptr;
}

obj *make_spec (obj *func) {
  obj_spec *ptr;

  num_nodes++;
  ptr = (obj_spec *) CollectMalloc (sizeof (obj_spec), 1);
  ptr->type = SPEC;
  ptr->func = func;
  return (obj *) ptr;
}

/* misc routines */

void error (char *message) {
  fprintf (stderr, "%s\n", message);
  exit (0);
}

void error_object (char *message, obj *sexp) {
  fprintf (stderr, "%s", message);
  write (stderr, sexp);
  fprintf (stderr, "\n");
  exit (0);
}

obj *lookup (char *name) {
  obj *ptr;

  for (ptr = names_head; ptr != nil; ptr = cdr(ptr))
    if (!strcmp (name, sym_name(car(ptr)))) break;

  if (ptr != nil)
    return car(ptr);
  else {
    ptr = make_sym (strdup(name));
    names_head = cons (ptr, names_head);
    return ptr;
  }
}

/* variable binding */

/* return an environment obtained by extending env */
obj *extend_env (obj *syms, obj *vals, obj *env) {
  if (vals == nil && syms == nil)
    return env;
  else if (vals != nil && syms == nil)
    error ("Binding error: more values than variables.");
  else if (vals == nil && syms != nil)
    error ("Binding error: more variables than values.");
  
  return cons (cons (car(syms), car(vals)),
	extend_env (cdr(syms), cdr(vals), env));
}

/* add sym to tail of top_env; bind sym to val */
void define (obj *sym, obj *val) {
  obj *ptr;

  for (ptr = top_env; cdr(ptr) != nil; ptr = cdr(ptr));
  cdr(ptr) = cons (cons (sym, val), nil);  
}

/* return pair in alist whose car is key, or nil if no such pair exists */
obj *assoc (obj *key, obj *alist) {
  obj *ptr;

  for (ptr = alist; ptr != nil; ptr = cdr(ptr))
    if (car(car(ptr)) == key) break;
  
  return (ptr == nil) ? nil : car(ptr);
}

/* read an s-expression */

#define ATOM 0

FILE *input_file;
obj *tokenval;

int next_token () {
  char buffer[128];
  int buffer_index;
  int ch;

  buffer_index = 0;

  // Read character and remove comments (;)
  do {
    ch = getc (input_file);
    if (ch == ';') {
      do ch = getc (input_file);
      while (ch != '\n' && ch != EOF);
    }
  } while(isspace (ch) && ch != EOF);

  if (ch == EOF)
    exit (0);
  else if (ch == ':')
    return getc (input_file);
  else if (strchr ("()\'", ch))
    return ch;
  else if (ch == '.') {
    ch = getc (input_file);
    if (strchr ("()\'", ch) || isspace (ch)) {
      ungetc (ch, input_file);
      return '.';
    }
    buffer[buffer_index++] = '.';
  }
  
  do {
    buffer[buffer_index++] = ch;
    ch = getc (input_file);
    if (ch == EOF)
      exit (0);
    else if (strchr ("()\'", ch) || isspace (ch)) {
      ungetc (ch, input_file);
      buffer[buffer_index++] = '\0';
      if (buffer[strspn(buffer, "0123456789")] == '\0')
        tokenval = make_int(atoi(buffer));
      else
        tokenval = lookup (buffer);
      return ATOM;
    }
  } while (1);
}

obj *read_sexp (int token) {
  if (token == '\'')
    return cons(s_quote, cons(read_sexp (next_token ()), nil));
  else if (token == '(')
    return read_list (next_token ());
  else if (token == ATOM)
    return tokenval;

  error ("Syntax error in s-expression.");
}

obj *read_list (int token) {
  obj *tmp;
  
  if (token == ')')
    return nil;
  else if (token == '.') {
    tmp = read_sexp (next_token ());
    if (next_token() == ')')
      return tmp;
    else
      error ("Syntax error in improper list.");
  }
  
  tmp = read_sexp (token);
  return cons (tmp, read_list (next_token ()));
}

obj *read () {
  int token;
  
  token = next_token ();

  switch (token) {
  case 'e':
    return cons (s_quote, cons (top_env, nil));
  case 'n':
    return cons (s_quote, cons (names_head, nil));
  case 'm':
    return make_int(num_nodes);
  case 'q':
    exit (0);
  case 't':
    flag_trace ^= 1;
    return (make_int(flag_trace));
  default:
    return read_sexp (token);
  }
}

/* print an s-expression */

void write (FILE *output, obj *sexp) {
  switch (sexp->type) {
  case CONS:
    fprintf (output, "(");
    write (output, car(sexp));
    sexp = cdr(sexp);
    while (sexp->type == CONS) {
      fprintf (output, " ");
      write (output, car(sexp));
      sexp = cdr(sexp);
    }
    if (sexp == nil)
	fprintf (output, ")");
    else {
      fprintf (output, " . ");
      write (output, sexp);
      fprintf (output, ")");
    }
    break;
  case INT: 
    fprintf (output, "%d", int_value(sexp));
    break;
  case SYM:
    fprintf (output, "%s", sym_name(sexp));
    break;
  case PRIM:
    fprintf (output, "[primitive function]");
    break;
  case COMP:
    fprintf (output, "[compound function]");
    break;
  case SPEC:
    fprintf (output, "[special form]");
    break;
  case PORT:
    fprintf (output, "[port:0x%x]", port_name(sexp));
    break;
  case TIMER:
    fprintf (output, "[timer]");
    break;
  default:
    error ("Cannot print object: invalid type.");
  }
}

/* evaluate an s-expression */

obj *eval (obj *sexp, obj *env) {
  obj *tmp;
  
  switch (sexp->type) {
  case CONS:
    tmp = eval (car(sexp), env);
    if (tmp->type == SPEC)
      return apply (spec_func(tmp), cons (cdr(sexp), cons (env, nil)));
    else
      return apply (tmp, eval_list (cdr(sexp), env));
  case SYM:
    tmp = assoc (sexp, env);
    if (tmp == nil) tmp = assoc (sexp, top_env);
    if (tmp == nil)
      error_object ("Unbound variable: ", sexp);
    else
      return cdr(tmp);
  case INT:
  case PRIM:
  case COMP:
  case SPEC:
  case PORT:
  case TIMER:
    return sexp;
  default:
    error ("Cannot evaluate object: invalid type.");
  }
}

obj *eval_list (obj *lst, obj *env) {
  if (lst == nil)
    return nil;
  else
    return cons(eval (car(lst), env), eval_list (cdr(lst), env));
}

void trace_lambda (obj *func, obj *args) {
  fprintf (stdout, "Evaluate ");
  write (stdout, comp_code(func));
  fprintf (stdout, "\n");
  fprintf (stdout, "in environment ");
  if (comp_syms(func)->type == CONS || comp_syms(func) == nil)
    write (stdout, extend_env (comp_syms(func), args, comp_env(func)));
  else
    write (stdout, cons (cons (comp_syms(func), args), comp_env(func)));
  fprintf (stdout, "\n");
}

obj *apply (obj *func, obj *args) {
  switch (func->type) {
  case PRIM:
    return prim_cfunc(func)(args);
  case COMP:
    if (flag_trace) trace_lambda (func, args);
    if (comp_syms(func)->type == CONS || comp_syms(func) == nil)
      return eval (comp_code(func),
		   extend_env (comp_syms(func), args, comp_env(func)));
    else if (comp_syms(func)->type == SYM)
      return eval (comp_code(func),
		   cons (cons (comp_syms(func), args), comp_env(func)));
    else
      error ("Error: invalid argument name(s) in lambda expression.");
  case TIMER:
	  return make_int(getTime());
  default:
    error_object ("Not applicable: ", func);
  }
}

/* primitive functions used by special forms */

obj *prim_quote (obj *args) {
  obj *sexp = car(args);

  return car(sexp);
}

obj *prim_lambda (obj *args) {
  obj *sexp = car(args), *env = second(args);

  return make_comp (second(sexp), car(sexp), env);
}

obj *prim_if (obj *args) {
  obj *sexp = car(args), *env = second(args);

  if (eval (car(sexp), env) != nil)
    return eval (second(sexp), env);
  else
    return eval (third(sexp), env);
}

obj *prim_define (obj *args) {
  obj *sexp = car(args), *env = second(args);
  obj *tmp;

  if (car(sexp)->type != SYM)
    error_object ("Not a symbol: ", car(sexp));
  tmp = assoc (car(sexp), top_env);
  if (tmp == nil)
    define (car(sexp), eval (second(sexp), env));
  else
    cdr(tmp) = eval (second(sexp), env);
  return car(sexp);
}

/* primitive functions */

obj *prim_pout (obj *args) {
	pout(port_name(car(args)),int_value(second(args)));
    return second(args);
}

obj *prim_pin (obj *args) {
	return make_int(pin(port_name(car(args))));
}

obj *prim_peq (obj *args) {
    return make_int(peq(port_name(car(args)),int_value(second(args))));
}

obj *prim_pne (obj *args) {
    return make_int(pne(port_name(car(args)),int_value(second(args))));
}

obj *prim_after (obj *args) {
	timeAfter(int_value(car(args)));
	return second(args);
}

obj *prim_sum (obj *args) {
  int sum;

  for (sum = 0; args != nil; args = cdr(args)) sum += int_value(car(args));
  return make_int(sum);
}

obj *prim_sub (obj *args) {
  int sum;

  if (args == nil) return make_int(0);
  sum = int_value(car(args));
  for (args = cdr(args); args != nil; args = cdr(args))
    sum -= int_value(car(args));
  return make_int(sum);
}

obj *prim_div (obj *args) {

  return make_int(int_value(car(args)) / int_value(second(args)));
}

obj *prim_mod (obj *args) {

  return make_int(int_value(car(args)) % int_value(second(args)));
}

obj *prim_prod (obj *args) {
  int prod;

  for (prod = 1; args != nil; args = cdr(args)) prod *= int_value(car(args));
  return make_int(prod);
}

obj *prim_numeq (obj *args) {
  return int_value(car(args)) == int_value(second(args)) ? tee : nil;
}

obj *prim_car (obj *args) {
  return car(car(args));
}

obj *prim_cdr (obj *args) {
  return cdr(car(args));
}

obj *prim_cons (obj *args) {
  return cons(car(args), second(args));
}

obj *prim_eval (obj *args) {
  return eval (car(args), second(args));
}

obj *prim_apply (obj *args) {
  return apply (car(args), second(args));
}

obj *prim_assoc (obj *args) {
  return assoc (car(args), second(args));
}

obj *prim_eqvp (obj *args) {
  if (car(args)->type == INT && second(args)->type == INT)
    return (int_value(car(args)) == int_value(second(args))) ? tee : nil;
  else
    return (car(args) == second(args)) ? tee : nil;
}

obj *prim_type_of (obj *args) {
  return make_int(car(args)->type);
}

obj *prim_special (obj *args) {
  return make_spec (car(args));
}

/* initialization */

void initialize () {

  /* define nil, initialize environment and names list */

  nil = make_sym ("()");
  names_head = nil;
  top_env = cons (cons (nil, nil), nil);

  /* #t and #f */

  define (lookup ("#f"), nil);
  tee = lookup ("#t");
  define (tee, tee);

  /* primitive functions */

  define (lookup ("+"), make_prim (prim_sum));
  define (lookup ("-"), make_prim (prim_sub));
  define (lookup ("*"), make_prim (prim_prod));
  define (lookup ("/"), make_prim (prim_div));
  define (lookup ("%"), make_prim (prim_mod));
  define (lookup ("="), make_prim (prim_numeq));
  define (lookup ("car"),  make_prim (prim_car));
  define (lookup ("cdr"),  make_prim (prim_cdr));
  define (lookup ("cons"), make_prim (prim_cons));
  define (lookup ("eval"), make_prim (prim_eval));
  define (lookup ("apply"), make_prim (prim_apply));
  define (lookup ("assoc"), make_prim (prim_assoc));
  define (lookup ("eqv?"), make_prim (prim_eqvp));
  define (lookup ("type-of"), make_prim (prim_type_of));
  define (lookup ("special"), make_prim (prim_special));
  define (lookup ("pout"),  make_prim (prim_pout));
  define (lookup ("pin"),  make_prim (prim_pin));
  define (lookup ("peq"),  make_prim (prim_peq));
  define (lookup ("pne"),  make_prim (prim_pne));
  define (lookup ("after"),  make_prim (prim_after));
  
  /* XMOS resources */
  if(get_core_id() == 0)
  {
	  define (lookup ("clockselr"), make_port (0x10400));
	  define (lookup ("clockselg"), make_port (0x10600));
	  define (lookup ("clockled0"), make_port (0x40000));
	  define (lookup ("clockled1"), make_port (0x40100));
	  define (lookup ("clockled2"), make_port (0x40200));
	  define (lookup ("button"),    make_port (0x40300));
	  define (lookup ("buttonled"), make_port (0x80300));
  }
  define (lookup ("timer"), make_timer());
  define (lookup ("core"), make_int (get_core_id()));
  
  /* special forms */

  define (lookup ("quote"), make_spec (make_prim (prim_quote)));
  define (lookup ("lambda"), make_spec (make_prim (prim_lambda)));
  define (lookup ("if"), make_spec (make_prim (prim_if)));
  define (lookup ("define"), make_spec (make_prim (prim_define)));
  
  s_quote = lookup ("quote");
}

/* read-eval-print loop */

int lisp () {
  initialize ();
  
  input_file = stdin;

  CollectRoot(&top_env);
  CollectRoot(&names_head);
  
  do {
    printf ("--> ");
    write (stdout, eval (read (), nil));
    printf ("\n");
    num_nodes = CollectGarbage();
  } while (1);

  return 0;
}
